//
//  NSEntityDescription+QuantumKitAdditions.m
//  QuantumKit
//
//  Created by Logan Collins on 7/9/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "NSEntityDescription+QuantumKitAdditions.h"

#import "QNKPersistentSyncCoordinator.h"
#import "NSPropertyDescription+QuantumKitAdditions.h"
#import <objc/runtime.h>


@implementation NSEntityDescription (QuantumKitAdditions)

static NSString * const QNKIsSyncableKey = @"com.quantumkit.syncable";

- (NSNumber *)qnk_recursiveSyncable {
    NSNumber *result = nil;
    id value = [[self userInfo] objectForKey:QNKEntityIsSyncableKey];
    if ([value isKindOfClass:[NSNumber class]]) {
        result = value;
    }
    else if ([value isKindOfClass:[NSString class]]) {
        if ([[value lowercaseString] isEqualToString:@"true"] || [[value lowercaseString] isEqualToString:@"yes"] || [[value lowercaseString] isEqualToString:@"1"]) {
            result = @YES;
        }
        else {
            result = @NO;
        }
    }
    
    if (result == nil) {
        // Only recurse if the result is nil, not false, as subentities can opt-out of syncing inherited from a superentity
        result = [[self superentity] qnk_recursiveSyncable];
    }
    
    return result;
}

- (BOOL)qnk_isSyncable {
    NSNumber *value = objc_getAssociatedObject(self, (__bridge void *)QNKIsSyncableKey);
    if (value == nil) {
        value = [self qnk_recursiveSyncable];
        if (value == nil) {
            value = [NSNumber numberWithBool:NO];
        }
        objc_setAssociatedObject(self, (__bridge void *)QNKIsSyncableKey, value, OBJC_ASSOCIATION_COPY);
    }
    return [value boolValue];
}

static NSString * const QNKSyncableIdentifierAttributeNameKey = @"com.quantumkit.identifier-attribute";

- (NSString *)qnk_recursiveSyncableIdentifierAttributeName {
    NSDictionary *userInfo = [self userInfo];
    NSString *value = userInfo[QNKEntityIdentifierAttributeNameKey];
    if (value == nil) {
        value = [[self superentity] qnk_recursiveSyncableIdentifierAttributeName];
    }
    return value;
}

- (NSString *)qnk_syncableIdentifierAttributeName {
    NSString *attributeName = objc_getAssociatedObject(self, (__bridge void *)QNKSyncableIdentifierAttributeNameKey);
    if (attributeName == nil) {
        attributeName = [self qnk_recursiveSyncableIdentifierAttributeName];
        if (attributeName == nil) {
            attributeName = @"identifier";
        }
        
        NSPropertyDescription *property = [[self propertiesByName] objectForKey:attributeName];
        if (property != nil && [property isKindOfClass:[NSAttributeDescription class]]) {
            if (![(NSAttributeDescription *)property attributeType] == NSStringAttributeType) {
                @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Identifier property '%@' of entity '%@' is not a string attribute.", attributeName, self] userInfo:nil];
            }
        }
        else if (property != nil) {
            @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Identifier property '%@' of entity '%@' is not a string attribute.", attributeName, self] userInfo:nil];
        }
        else {
            @throw [NSException exceptionWithName:NSInvalidArgumentException reason:[NSString stringWithFormat:@"Identifier property '%@' of entity '%@' is not defined.", attributeName, self] userInfo:nil];
        }
        
        objc_setAssociatedObject(self, (__bridge void *)QNKSyncableIdentifierAttributeNameKey, attributeName, OBJC_ASSOCIATION_COPY);
    }
    return attributeName;
}

static NSString * const QNKSyncablePropertiesByNameKey = @"QNKSyncablePropertiesByName";

- (NSDictionary *)qnk_syncablePropertiesByName {
    NSDictionary *syncablePropertiesByName = nil;
    @synchronized(self) {
        syncablePropertiesByName = objc_getAssociatedObject(self, (__bridge void *)QNKSyncablePropertiesByNameKey);
        if (syncablePropertiesByName == nil) {
            NSDictionary *properties = [self propertiesByName];
            NSMutableDictionary *newSyncablePropertiesByName = [NSMutableDictionary dictionaryWithCapacity:[properties count]];
            
            for (NSString *name in properties) {
                NSPropertyDescription *property = properties[name];
                
                BOOL isSyncable = [property qnk_isSyncable];
                if (isSyncable) {
                    newSyncablePropertiesByName[name] = property;
                }
            }
            
            syncablePropertiesByName = [newSyncablePropertiesByName mutableCopy];
            objc_setAssociatedObject(self, (__bridge void *)QNKSyncablePropertiesByNameKey, syncablePropertiesByName, OBJC_ASSOCIATION_COPY);
        }
    }
    return syncablePropertiesByName;
}

@end
