//
//  NSManagedObject+QuantumKitAdditions.m
//  QuantumKit
//
//  Created by Logan Collins on 7/9/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "NSManagedObject+QuantumKitAdditions.h"

#import "NSEntityDescription+QuantumKitAdditions.h"
#import "QuantumKitErrors.h"


@implementation NSManagedObject (QuantumKitAdditions)

- (NSString *)qnk_syncableIdentifier {
    NSString *attributeName = [[self entity] qnk_syncableIdentifierAttributeName];
    id value = [self valueForKey:attributeName];
    if (value == nil) {
        //@throw [NSException exceptionWithName:NSInvalidArgumentException reason:@"Syncable identifier cannot be nil" userInfo:nil];
    }
    return value;
}

- (void)qnk_setSyncableIdentifier:(NSString *)identifier {
    NSString *attributeName = [[self entity] qnk_syncableIdentifierAttributeName];
    [self setValue:identifier forKey:attributeName];
}

- (NSDictionary *)qnk_objectsWithIdentifiers:(NSArray *)identifiers entityName:(NSString *)entityName {
    __block NSMutableDictionary *objects = nil;
    
    NSManagedObjectContext *context = [self managedObjectContext];
    NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:entityName];
    
    NSString *syncableIdentifierAttributeName = [[self entity] qnk_syncableIdentifierAttributeName];
    request.predicate = [NSPredicate predicateWithFormat:@"%K IN %@", syncableIdentifierAttributeName, identifiers];
    
    NSArray *results = [context executeFetchRequest:request error:nil];
    objects = [NSMutableDictionary dictionaryWithCapacity:[results count]];
    for (NSManagedObject *object in results) {
        objects[[object qnk_syncableIdentifier]] = object;
    }
    
    return objects;
}

- (id)qnk_syncableValueForKey:(NSString *)key error:(NSError **)outError {
    id property = nil;
    NSError *error = nil;
    
    NSEntityDescription *entity = [self entity];
    id objectProperty = [[entity propertiesByName] objectForKey:key];
    if (objectProperty == nil) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                  NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                  [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to an attribute of entity '%@'", nil), key, [entity name]], NSLocalizedFailureReasonErrorKey,
                                  nil];
        error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKUnsupportedPropertyMappingError userInfo:userInfo];
    }
    
    if ([objectProperty isKindOfClass:[NSAttributeDescription class]]) {
        NSAttributeType type = [objectProperty attributeType];
        id value = [self valueForKey:key];
        
        if (type == NSInteger16AttributeType
            || type == NSInteger32AttributeType
            || type == NSInteger64AttributeType) {
            
        }
        else if (type == NSFloatAttributeType
                 || type == NSDoubleAttributeType) {
            
        }
        else if (type == NSDecimalAttributeType) {
            // Convert decimal numbers to strings
            static NSNumberFormatter *numberFormatter = nil;
            static dispatch_once_t onceToken;
            dispatch_once(&onceToken, ^{
                numberFormatter = [[NSNumberFormatter alloc] init];
                [numberFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
                [numberFormatter setGeneratesDecimalNumbers:YES];
            });
            value = [numberFormatter stringFromNumber:(NSDecimalNumber *)value];
        }
        else if (type == NSBooleanAttributeType) {
            
        }
        else if (type == NSStringAttributeType) {
            
        }
        else if (type == NSDateAttributeType) {
            
        }
        else if (type == NSBinaryDataAttributeType) {
            
        }
        else if (type == NSTransformableAttributeType) {
            // Convert transformable types to data
            NSString *valueTransformerName = [objectProperty valueTransformerName];
            if (value == nil) {
                // no-op
            }
            else if (valueTransformerName == nil) {
                value = [NSKeyedArchiver archivedDataWithRootObject:value];
            }
            else {
                NSValueTransformer *transformer = [NSValueTransformer valueTransformerForName:valueTransformerName];
                value = [transformer transformedValue:value];
            }
        }
        else {
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                      [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to an attribute of entity '%@'", nil), key, [entity name]], NSLocalizedFailureReasonErrorKey,
                                      nil];
            error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKUnsupportedPropertyMappingError userInfo:userInfo];
        }
        
        property = value;
    }
    else if ([objectProperty isKindOfClass:[NSRelationshipDescription class]]) {
        NSEntityDescription *destinationEntity = [objectProperty destinationEntity];
        if ([destinationEntity qnk_isSyncable]) {
            if ([objectProperty isToMany]) {
                id value = [self valueForKey:key];
                if ([objectProperty isOrdered]) {
                    NSOrderedSet *relatedObjects = value;
                    NSMutableArray *array = [NSMutableArray arrayWithCapacity:[relatedObjects count]];
                    for (NSManagedObject *relatedObject in relatedObjects) {
                        [array addObject:[relatedObject qnk_syncableIdentifier]];
                    }
                    property = array;
                }
                else {
                    NSSet *relatedObjects = value;
                    NSMutableArray *array = [NSMutableArray arrayWithCapacity:[relatedObjects count]];
                    for (NSManagedObject *relatedObject in relatedObjects) {
                        [array addObject:[relatedObject qnk_syncableIdentifier]];
                    }
                    property = array;
                }
            }
            else {
                NSManagedObject *relatedObject = [self valueForKey:key];
                if (relatedObject != nil) {
                    property = @[[relatedObject qnk_syncableIdentifier]];
                }
            }
        }
        else {
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                      [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to a relationship of entity '%@'", nil), key, [entity name]], NSLocalizedFailureReasonErrorKey,
                                      nil];
            error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKUnsupportedPropertyMappingError userInfo:userInfo];
        }
    }
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return property;
}

- (BOOL)qnk_setSyncableValue:(id)value forKey:(NSString *)key error:(NSError **)outError {
    BOOL success = NO;
    NSError *error = nil;
    
    NSEntityDescription *entity = [self entity];
    id objectProperty = [[entity propertiesByName] objectForKey:key];
    if (objectProperty == nil) {
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                  NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                  [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to an attribute of entity '%@'", nil), key, [entity name]], NSLocalizedFailureReasonErrorKey,
                                  nil];
        error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKUnsupportedPropertyMappingError userInfo:userInfo];
        success = NO;
    }
    
    if ([objectProperty isKindOfClass:[NSAttributeDescription class]]) {
        // Attribute
        NSAttributeType type = [objectProperty attributeType];
        
        id propertyValue = nil;
        
        BOOL valid = NO;
        
        if (type == NSInteger16AttributeType
            || type == NSInteger32AttributeType
            || type == NSInteger64AttributeType
            || type == NSFloatAttributeType
            || type == NSDoubleAttributeType
            || type == NSBooleanAttributeType) {
            if (value == nil || [value isKindOfClass:[NSNumber class]]) {
                propertyValue = value;
                valid = YES;
            }
        }
        else if (type == NSDecimalAttributeType) {
            if (value == nil || [value isKindOfClass:[NSString class]]) {
                static NSNumberFormatter *numberFormatter = nil;
                static dispatch_once_t onceToken;
                dispatch_once(&onceToken, ^{
                    numberFormatter = [[NSNumberFormatter alloc] init];
                    [numberFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
                    [numberFormatter setGeneratesDecimalNumbers:YES];
                });
                propertyValue = [numberFormatter numberFromString:value];
                valid = YES;
            }
        }
        else if (type == NSStringAttributeType) {
            if (value == nil || [value isKindOfClass:[NSString class]]) {
                propertyValue = value;
                valid = YES;
            }
        }
        else if (type == NSDateAttributeType) {
            if (value == nil || [value isKindOfClass:[NSDate class]]) {
                propertyValue = value;
                valid = YES;
            }
        }
        else if (type == NSBinaryDataAttributeType) {
            if (value == nil || [value isKindOfClass:[NSData class]]) {
                propertyValue = value;
                valid = YES;
            }
        }
        else if (type == NSTransformableAttributeType) {
            if (value == nil || [value isKindOfClass:[NSData class]]) {
                NSString *valueTransformerName = [(NSAttributeDescription *)objectProperty valueTransformerName];
                if (value == nil) {
                    propertyValue = value;
                }
                else if (valueTransformerName == nil) {
                    propertyValue = [NSKeyedUnarchiver unarchiveObjectWithData:value];
                }
                else {
                    NSValueTransformer *transformer = [NSValueTransformer valueTransformerForName:valueTransformerName];
                    propertyValue = [transformer reverseTransformedValue:value];
                }
                valid = YES;
            }
        }
        
        if (valid) {
            if (propertyValue == nil && ![objectProperty isOptional] && [(NSAttributeDescription *)objectProperty defaultValue] != nil) {
                // Reset to default value
                propertyValue = [(NSAttributeDescription *)objectProperty defaultValue];
            }
            
            if (propertyValue == nil && ![objectProperty isOptional]) {
                NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                          NSLocalizedString(@"Attribute cannot be empty", nil), NSLocalizedDescriptionKey,
                                          [NSString stringWithFormat:NSLocalizedString(@"Required attribute '%@' of entity '%@' cannot be empty.", nil), key, [entity name]], NSLocalizedDescriptionKey,
                                          nil];
                error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKMissingRequiredValueError userInfo:userInfo];
                success = NO;
            }
            else {
                // Validate and set
                if ([self validateValue:&propertyValue forKey:key error:&error]) {
                    [self setValue:propertyValue forKey:key];
                    success = YES;
                }
            }
        }
        else {
            // Unsupported attribute type
            NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                      NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                      [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map attribute '%@' to an attribute of entity '%@'", nil), key, [entity name]], NSLocalizedFailureReasonErrorKey,
                                      nil];
            error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKUnsupportedPropertyMappingError userInfo:userInfo];
            success = NO;
        }
    }
    else if ([objectProperty isKindOfClass:[NSRelationshipDescription class]] && [value isKindOfClass:[NSArray class]]) {
        // Relationship
        NSEntityDescription *destinationEntity = [objectProperty destinationEntity];
        
        NSArray *relatedIdentifiers = value;
        NSDictionary *relatedObjects = [self qnk_objectsWithIdentifiers:relatedIdentifiers entityName:[destinationEntity name]];
        
        id newValue = nil;
        
        if ([(NSRelationshipDescription *)objectProperty isToMany]) {
            if ([(NSRelationshipDescription *)objectProperty isOrdered]) {
                NSMutableOrderedSet *set = [NSMutableOrderedSet orderedSetWithCapacity:[relatedIdentifiers count]];
                for (NSString *relatedIdentifier in relatedIdentifiers) {
                    NSManagedObject *relatedObject = relatedObjects[relatedIdentifier];
                    if (relatedObject != nil) {
                        [set addObject:relatedObject];
                    }
                }
                newValue = set;
            }
            else {
                NSMutableSet *set = [NSMutableSet setWithCapacity:[relatedIdentifiers count]];
                for (NSString *relatedIdentifier in relatedIdentifiers) {
                    NSManagedObject *relatedObject = relatedObjects[relatedIdentifier];
                    if (relatedObject != nil) {
                        [set addObject:relatedObject];
                    }
                }
                newValue = set;
            }
        }
        else {
            if ([relatedIdentifiers count] > 0) {
                NSString *relatedIdentifier = relatedIdentifiers[0];
                NSManagedObject *relatedObject = relatedObjects[relatedIdentifier];
                newValue = relatedObject;
            }
        }
        
        if ([self validateValue:&newValue forKey:key error:&error]) {
            [self setValue:newValue forKey:key];
            success = YES;
        }
    }
    else {
        // Unknown
        NSDictionary *userInfo = [NSDictionary dictionaryWithObjectsAndKeys:
                                  NSLocalizedString(@"Unsupported property mapping", nil), NSLocalizedDescriptionKey,
                                  [NSString stringWithFormat:NSLocalizedString(@"Unable to automatically map property '%@' to a property of entity '%@'", nil), key, [entity name]], NSLocalizedFailureReasonErrorKey,
                                  nil];
        error = [NSError errorWithDomain:QNKQuantumKitErrorDomain code:QNKUnsupportedPropertyMappingError userInfo:userInfo];
        success = NO;
    }
    
    if (outError != NULL) {
        *outError = error;
    }
    
    return success;
}

@end
