//
//  QNKPersistentSyncCoordinator.h
//  QuantumKit
//
//  Created by Logan Collins on 7/9/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <CoreData/CoreData.h>
#import <QuantumKit/QuantumKitDefines.h>


@protocol QNKPersistentSyncCoordinatorDelegate;
@class DBDatastore;


/*!
 * @class QNKPersistentCoordinator
 * @abstract Automatically coordinates sync for Core Data stores
 *
 * @discussion
 * A persistent sync coordinator works in tandem with a primary managed
 * object context to monitor for changes in the object graph, relaying these
 * changes to a sync service, and applying changes received from the sync
 * service automatically.
 *
 * To support syncing, a persistent sync coordinator must be created for
 * a managed object context, and any managed object within the store that will
 * be synced must be marked as syncable using the userInfo keys defined below.
 */
@interface QNKPersistentSyncCoordinator : NSObject

- (id)initWithDatastore:(DBDatastore *)datastore managedObjectContext:(NSManagedObjectContext *)moc;

@property (strong, readonly) DBDatastore *datastore;
@property (strong, readonly) NSManagedObjectContext *managedObjectContext;

@property (unsafe_unretained) id <QNKPersistentSyncCoordinatorDelegate> delegate;

@property (readonly, getter=isSyncEnabled) BOOL syncEnabled;
- (void)disableSync;
- (void)enableSync;

- (void)reset;

@end


/*!
 * @protocol QNKPersistentSyncCoordinatorDelegate
 * @abstract Adopted by a persistent sync coordinator delegate
 */
@protocol QNKPersistentSyncCoordinatorDelegate <NSObject>

@optional

// Push
- (NSDictionary *)persistentSyncCoordinator:(QNKPersistentSyncCoordinator *)coordinator willPushRecord:(NSDictionary *)record forManagedObject:(NSManagedObject *)managedObject;
- (BOOL)persistentSyncCoordinator:(QNKPersistentSyncCoordinator *)coordinator shouldDeleteRecordWithIdentifier:(NSString *)identifier forManagedObject:(NSManagedObject *)managedObject;

// Pull
- (NSDictionary *)persistentSyncCoordinator:(QNKPersistentSyncCoordinator *)coordinator willApplyRecord:(NSDictionary *)record toManagedObject:(NSManagedObject *)managedObject;
- (void)persistentSyncCoordinator:(QNKPersistentSyncCoordinator *)coordinator didApplyRecord:(NSDictionary *)record toManagedObject:(NSManagedObject *)managedObject;

@end


/*!
 * @group Entity User Info Keys
 */

/*!
 * @constant QNKEntityIsSyncableKey
 * @abstract Informs a persistent sync coordinator that an entity is syncable.
 *
 * @discussion
 * This key should be set to a truth value (NSNumber or NSString) in an NSEntityDescription's userInfo dictionary.
 */
QUANTUMKIT_EXTERN NSString * const QNKEntityIsSyncableKey; // @"com.quantumkit.syncable";

/*!
 * @constant QNKEntityIdentifierAttributeNameKey
 * @abstract Defines the name of the unique, syncable identifier attribute for an entity class
 *
 * @discussion
 * This key should be set to the name of a modelled attribute (NSString) in every root NSEntityDescription's userInfo dictionary.
 *
 * If no identifier attribute name is specified, a default value of "identifier" is used.
 * If the entity does not define a modelled string attribute with this name an exception is raised when a sync session begins.
 *
 * A subentity automatically becomes syncable if its superentity defines this key. To disable, set this key to a false truth value in the subentity.
 */
QUANTUMKIT_EXTERN NSString * const QNKEntityIdentifierAttributeNameKey; // @"com.quantumkit.identifier-attribute";


/*!
 * @group Property User Info Keys
 */

/*!
 * @constant QNKPropertyIsSyncableKey
 * @abstract Informs a persistent sync coordinator that a property is syncable.
 *
 * @discussion
 * This key should be set to a truth value (NSNumber or NSString) in an NSPropertyDescription's userInfo dictionary.
 * By default, all attributes are ignored and must be explicitly marked for syncing.
 */
QUANTUMKIT_EXTERN NSString * const QNKPropertyIsSyncableKey; // @"com.quantumkit.syncable";


