//
//  QNKPersistentSyncCoordinator.m
//  QuantumKit
//
//  Created by Logan Collins on 7/9/13.
//  Copyright (c) 2013 Sunflower Softworks. All rights reserved.
//

#import "QNKPersistentSyncCoordinator.h"
#import "QNKPersistentSyncCoordinator_Private.h"

#import "QNKPersistentSyncChange.h"
#import "QNKPersistentSyncEntity.h"

#import "NSManagedObject+QuantumKitAdditions.h"
#import "NSEntityDescription+QuantumKitAdditions.h"

#import "QuantumKitErrors.h"
#import "QNKDebug.h"

#import <Dropbox/Dropbox.h>


NSString * const QNKEntityIsSyncableKey = @"com.quantumkit.syncable";
NSString * const QNKEntityIdentifierAttributeNameKey = @"com.quantumkit.identifier-attribute";

NSString * const QNKPropertyIsSyncableKey = @"com.quantumkit.syncable";


@implementation QNKPersistentSyncCoordinator {
    NSOperationQueue *_operationQueue;
	
	DBTable *_deviceTable;
	DBTable *_entityTable;
}

- (id)initWithDatastore:(DBDatastore *)datastore managedObjectContext:(NSManagedObjectContext *)moc {
    NSParameterAssert(datastore != nil);
    NSParameterAssert(moc != nil);
    
    self = [super init];
    if (self) {
        self.datastore = datastore;
        self.managedObjectContext = moc;
		
		_deviceTable = [datastore getTable:@"com.quantumkit.meta.Device"];
		_entityTable = [datastore getTable:@"com.quantumkit.meta.Entity"];
        
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(managedObjectContextWillSave:) name:NSManagedObjectContextWillSaveNotification object:moc];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(managedObjectContextDidSave:) name:NSManagedObjectContextDidSaveNotification object:moc];
        
        [datastore addObserver:self block:^{
            [self syncChanges];
        }];
    }
    return self;
}

- (void)dealloc {
    [self.datastore removeObserver:self];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}


#pragma mark -
#pragma mark Notifications

- (void)managedObjectContextWillSave:(NSNotification *)notification {
	
}

- (void)managedObjectContextDidSave:(NSNotification *)notification {
    NSManagedObjectContext *context = [notification object];
    
    NSSet *insertedObjects = [context insertedObjects];
    NSSet *updatedObjects = [context updatedObjects];
    NSSet *deletedObjects = [context deletedObjects];
    
    if ([insertedObjects count] > 0 || [updatedObjects count] > 0 || [deletedObjects count] > 0) {
        QNKDebugLog(@"Observing managed object changes: inserted=%@, updated=%@, deleted=%@", insertedObjects, updatedObjects, deletedObjects);
        
		for (id object in insertedObjects) {
			if ([[object entity] qnk_isSyncable]) {
				NSString *entityName = [[object entity] name];
				NSString *identifier = [object qnk_syncableIdentifier];
				if (identifier != nil) {
					DBTable *table = [self.datastore getTable:[NSString stringWithFormat:@"com.quantumkit.data.%@", entityName]];
					DBRecord *record = [table getOrInsertRecord:identifier fields:nil inserted:NULL error:NULL];
					
					// Sync all properties for inserted objects
                    NSSet *propertyNames = [NSSet setWithArray:[[entity qnk_syncablePropertiesByName] allKeys]];
                    
					// Collect the values
					NSMutableDictionary *recordValues = [NSMutableDictionary dictionaryWithCapacity:[propertyNames count]];
					
                    for (NSString *propertyName in propertyNames) {
                        NSError *error = nil;
                        id propertyValue = [object qnk_syncableValueForKey:propertyName error:&error];
                        if (propertyValue != nil) {
                            [recordValues setObject:propertyValue forKey:propertyName];
                        }
                        else if (error != nil) {
							
                        }
                    }
					
					// Inform the delegate
					if (delegateRespondsToWillPushRecord) {
						recordValues = [self.delegate persistentSyncCoordinator:self willPushRecord:recordValues forManagedObject:object];
					}
					
					// Apply the values
					for (NSString *key in recordValues) {
						id value = recordValues[key];
						[record setValue:value forKey:key];
					}
				}
			}
		}
		
		for (id object in updatedObjects) {
			if ([[object entity] qnk_isSyncable]) {
				NSString *identifier = [object qnk_syncableIdentifier];
				
				NSMutableSet *syncableAttributes = [NSMutableSet setWithArray:[[[object entity] qnk_syncablePropertiesByName] allKeys]];
				NSSet *changeKeys = [NSSet setWithArray:[[object changedValues] allKeys]];
				[syncableAttributes intersectSet:changeKeys];
				
				if (identifier != nil && [syncableAttributes count] > 0) {
					DBTable *table = [self.datastore getTable:[NSString stringWithFormat:@"com.quantumkit.data.%@", entityName]];
					DBRecord *record = [table getRecord:identifier error:NULL];
					
					// Sync all properties for inserted objects
                    NSSet *propertyNames = nil;
//					if (changedAttributes != nil) {
//						// Sync only changed objects
//						propertyNames = changedAttributes;
//					}
//					else {
//						// Sync all attributes
						propertyNames = [NSSet setWithArray:[[entity sk_syncablePropertiesByName] allKeys]];
//					}
                    
					// Collect the values
					NSMutableDictionary *recordValues = [NSMutableDictionary dictionaryWithCapacity:[propertyNames count]];
					
                    for (NSString *propertyName in propertyNames) {
                        NSError *error = nil;
                        id propertyValue = [object qnk_syncableValueForKey:propertyName error:&error];
                        if (propertyValue != nil) {
                            [recordValues setObject:propertyValue forKey:propertyName];
                        }
                        else if (error != nil) {
							
                        }
                    }
					
					// Inform the delegate
					if (delegateRespondsToWillPushRecord) {
						recordValues = [self.delegate persistentSyncCoordinator:self willPushRecord:recordValues forManagedObject:object];
					}
					
					// Apply the values
					for (NSString *key in recordValues) {
						id value = recordValues[key];
						[record setValue:value forKey:key];
					}
				}
			}
		}
		
		for (id object in deletedObjects) {
			if ([[object entity] qnk_isSyncable]) {
				NSString *identifier = [object qnk_syncableIdentifier];
				if (identifier != nil) {
					DBTable *table = [self.datastore getTable:[NSString stringWithFormat:@"com.quantumkit.data.%@", entityName]];
					DBRecord *record = [table getRecord:identifier error:NULL];
					
					BOOL shouldDelete = YES;
					
					// Inform the delegate
					if (delegateRespondsToShouldDeleteRecord) {
						shouldDelete = [self.delegate persistentSyncCoordinator:self shouldDeleteRecordWithIdentifier:identifier forManagedObject:object];
					}
					
					if (shouldDelete) {
						[record deleteRecord];
					}
				}
			}
		}
		
		[self syncChanges];
    }
}


#pragma mark -
#pragma mark Syncing

- (DBRecord *)deviceRecordWithIdentifier:(NSString *)identifier {
	
}

- (NSDictionary *)syncableEntitiesByName {
	NSArray *entities = self.managedObjectContext.persistentStoreCoordinator.managedObjectModel.entities;
	
	NSMutableDictionary *entitiesToSync = [NSMutableDictionary dictionaryWithCapacity:[entities count]];
	for (NSEntityDescription *entity in entities) {
		NSString *name = [entity name];
		if ([entity qnk_isSyncable] && ![entity isAbstract]) {
			entitiesToSync[name] = entity;
		}
	}
	
	DBError *error = nil;
	NSArray *results = [_entityTable query:@{} error:&error];
	NSMutableDictionary *resultsByName = [NSMutableDictionary dictionaryWithCapacity:[results count]];
	for (DBRecord *entityRecord in results) {
		NSString *name = [entityRecord objectForKey:@"name"];
		resultsByName[name] = results;
	}
	
	NSFetchRequest *request = [NSFetchRequest fetchRequestWithEntityName:@"Entity"];
	NSArray *results = [_metadataManagedObjectContext executeFetchRequest:request error:nil];
	NSMutableDictionary *resultsByName = [NSMutableDictionary dictionaryWithCapacity:[results count]];
	for (SKPersistentSyncEntity *entity in results) {
		resultsByName[entity.name] = entity;
	}
	
	for (NSString *entityName in entitiesToSync) {
		SKPersistentSyncEntity *entity = resultsByName[entityName];
		if (entity == nil) {
			entity = [NSEntityDescription insertNewObjectForEntityForName:@"Entity" inManagedObjectContext:metadataContext];
			entity.name = entityName;
		}
		entity.shouldPerformSlowSync = YES;
	}
	
	NSError *error = nil;
	if (![metadataContext save:&error]) {
		NSLog(@"Error while saving metadata context: %@", error);
	}
}

- (void)syncChanges {
	NSManagedObjectContext *syncContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];
	syncContext.parentContext = self.managedObjectContext;
	
	NSDictionary *changes = [self.datastore sync:NULL];
	for (NSString *identifier in changes) {
		DBRecord *record = changes[identifier];
		
	}
}

@end
